import yaml
import psycopg2
import MySQLdb as MySql
from os import path as path

from s_logger.S_Logger import S_Logger

PG_ERROR = psycopg2.Error


class SConnector(object):
    """ Connector - parent class """

    def __init__(self, group_name):
        self._logger = S_Logger().get_logger(__file__)

        try:
            """
                Specify the path to the file with the connection settings (credential.yaml).
                File must be protected from unauthorized access, as it will contain
                passwords and usernames that will be used to connect to databases.
            """

            with open(path.expanduser(path.join('db_credential.yaml'))) as cred_path:
                params = yaml.safe_load(cred_path)
                file_path = str(params['path_to_cred'])

            with open(path.expanduser(file_path)) as file:
                params = yaml.safe_load(file)
                self._url = str(params[group_name]['host'])
                self._port = str(params[group_name]['port'])
                self._db = str(params[group_name]['database'])
                self._user = str(params[group_name]['user'])
                self._pass = str(params[group_name]['passcode'])
        except yaml.YAMLError as ex:
            self._logger.critical('ERROR: {}'.format(str(ex)), exc_info=True)
        except KeyError as ex:
            self._logger.critical('ERROR: {}'.format(str(ex)), exc_info=True)

        self._connection = None
        self._cursor = None
        self._connector_name = None

    def _run_script(self, script):
        self._cursor.execute(script)

    def select(self, script):
        self._logger.debug('SELECT SQL: {}'.format(script))
        try:
            self._cursor.execute(script)
            return self._cursor.fetchall()
        except PG_ERROR as ex:
            self._logger.error('Msg: `select` {code}\n{error}'
                               .format(code=ex.pgcode, error=ex.pgerror),
                               exc_info=True)
            return None

    def create(self, script):
        self._logger.debug('CREATE SQL: {}'.format(script))
        try:
            self._run_script(script)
        except PG_ERROR as ex:
            self._logger.error('Msg: `create` {code}\n{error}'
                               .format(code=ex.pgcode, error=ex.pgerror),
                               exc_info=True)

    def insert(self, script):
        self._logger.debug('INSERT SQL: {}'.format(script))
        try:
            self._run_script(script)
        except PG_ERROR as ex:
            self._logger.error('Msg: `insert` {code}\n{error}'
                               .format(code=ex.pgcode, error=ex.pgerror),
                               exc_info=True)

    def update(self, script):
        self._logger.debug('UPDATE SQL: {}'.format(script))
        try:
            self._run_script(script)
        except PG_ERROR as ex:
            self._logger.error('Msg: `update` {code}\n{error}'
                               .format(code=ex.pgcode, error=ex.pgerror),
                               exc_info=True)

    def exec(self, script):
        self._logger.debug('EXEC SQL: {}'.format(script))
        try:
            self._run_script(script)
        except PG_ERROR as ex:
            self._logger.error('Msg: `exec` {code}\n{error}'
                               .format(code=ex.pgcode, error=ex.pgerror),
                               exc_info=True)

    def call(self, procedure, args):
        self._logger.debug('CALL PROCEDURE `{p}` ARGS {a}'.format(p=procedure, a=args))
        try:
            if args is None or len(args) == 0:
                self._cursor.callproc(procedure)
            else:
                self._cursor.callproc(procedure, args)
        except PG_ERROR as ex:
            self._logger.error('Msg: `call` {code}\n{error}'
                               .format(code=ex.pgcode, error=ex.pgerror),
                               exc_info=True)

    def rollback(self):
        self._logger.debug('ROLLBACK ACTION')
        try:
            self._connection.rollback()
        except PG_ERROR as ex:
            self._logger.error('Msg: `rollback` {code}\n{error}'
                               .format(code=ex.pgcode, error=ex.pgerror),
                               exc_info=True)

    def close(self):
        self._cursor.close()
        self._cursor = None

        self._connection.close()
        self._connection = None

        self._logger.info('Close connection')

    def __exit__(self, type, value, traceback):
        self.close()

    def __str__(self):
        return '{url}:{port}/{db} {user}'.format(
            url=self._url,
            port=self._port,
            db=self._db,
            user=self._user
        )


class MySQLConnector(SConnector):
    """ MySQL connector """
    def open(self):
        try:
            self._connection = MySql.Connect(user=self._user, password=self._pass, host=self._url)
        except RuntimeError as ex:
            self._connection = MySql.connect(self._url, self._user, self._pass)

        self._connection.autocommit = True
        self._cursor = self._connection.cursor()
        self._connector_name = 'MySQL'
        self._logger.info('MySQL connection opened')

    def __enter__(self):
        self.open()
        return self


class RedshiftConnector(SConnector):
    """ Redshift connector """
    def open(self):
        super()
        self._connection = psycopg2.connect(host=self._url,
                                            user=self._user,
                                            port=self._port,
                                            password=self._pass,
                                            dbname=self._db)
        self._connection.autocommit = True
        self._cursor = self._connection.cursor()
        self._connector_name = 'RedShift'
        self._logger.info('Redshift connection opened')

    def __enter__(self):
        self.open()
        return self
